==> nimble-miner-api-0.0.10.tar.gz: nimble-miner-api-0.0.10/nimble/stream.py <==
# The MIT License (MIT)
# Copyright © 2024 Nimble Labs Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import nimble

from starlette.responses import StreamingResponse as _StreamingResponse
from starlette.responses import Response
from starlette.types import Send, Receive, Scope
from typing import Callable, Awaitable, List
from pydantic import BaseModel
from abc import ABC, abstractmethod


class NBStreamingResponseModel(BaseModel):
    """
    NBStreamingResponseModel is a Pydantic model that encapsulates the token streamer callable for Pydantic validation.
    It is used within the StreamingSynapse class to create a NBStreamingResponse object, which is responsible for handling
    the streaming of tokens.

    The token streamer is a callable that takes a send function and returns an awaitable. It is responsible for generating
    the content of the streaming response, typically by processing tokens and sending them to the client.

    This model ensures that the token streamer conforms to the expected signature and provides a clear interface for
    passing the token streamer to the NBStreamingResponse class.

    Attributes:
        token_streamer: Callable[[Send], Awaitable[None]]
            The token streamer callable, which takes a send function (provided by the ASGI server) and returns an awaitable.
            It is responsible for generating the content of the streaming response.
    """

    token_streamer: Callable[[Send], Awaitable[None]]


class StreamingSynapse(nimble.Synapse, ABC):
    """
    The StreamingSynapse class is designed to be subclassed for handling streaming responses in the nimble network.
    It provides abstract methods that must be implemented by the subclass to deserialize, process streaming responses,
    and extract JSON data. It also includes a method to create a streaming response object.
    """

    class Config:
        validate_assignment = True

    class NBStreamingResponse(_StreamingResponse):
        """
        NBStreamingResponse is a specialized subclass of the Starlette StreamingResponse designed to handle the streaming
        of tokens within the nimble network. It is used internally by the StreamingSynapse class to manage the response
        streaming process, including sending headers and calling the token streamer provided by the subclass.

        This class is not intended to be directly instantiated or modified by developers subclassing StreamingSynapse.
        Instead, it is used by the create_streaming_response method to create a response object based on the token streamer
        provided by the subclass.

        Reference: https://www.starlette.io/responses/#streamingresponse
        """

        def __init__(self, model: NBStreamingResponseModel, **kwargs):
            """
            Initializes the NBStreamingResponse with the given token streamer model.

            Args:
                model: A NBStreamingResponseModel instance containing the token streamer callable, which is responsible
                    for generating the content of the response.
                **kwargs: Additional keyword arguments passed to the parent StreamingResponse class.
            """
            super().__init__(content=iter(()), **kwargs)
            self.token_streamer = model.token_streamer

        async def stream_response(self, send: Send):
            """
            Asynchronously streams the response by sending headers and calling the token streamer.

            This method is responsible for initiating the response by sending the appropriate headers, including the
            content type for event-streaming. It then calls the token streamer to generate the content and sends the
            response body to the client.

            Args:
                send: A callable to send the response, provided by the ASGI server.
            """
            headers = [
                (b"content-type", b"text/event-stream")
            ] + self.raw_headers

            await send(
                {
                    "type": "http.response.start",
                    "status": 200,
                    "headers": headers,
                }
            )

            await self.token_streamer(send)

            await send(
                {"type": "http.response.body", "body": b"", "more_body": False}
            )

        async def __call__(self, scope: Scope, receive: Receive, send: Send):
            """
            Asynchronously calls the stream_response method, allowing the NBStreamingResponse object to be used as an ASGI
            application.

            This method is part of the ASGI interface and is called by the ASGI server to handle the request and send the
            response. It delegates to the stream_response method to perform the actual streaming process.

            Args:
                scope: The scope of the request, containing information about the client, server, and request itself.
                receive: A callable to receive the request, provided by the ASGI server.
                send: A callable to send the response, provided by the ASGI server.
            """
            await self.stream_response(send)

    @abstractmethod
    async def process_streaming_response(self, response: Response):
        """
        Abstract method that must be implemented by the subclass.
        This method should provide logic to handle the streaming response, such as parsing and accumulating data.
        It is called as the response is being streamed from the network, and should be implemented to handle the specific
        streaming data format and requirements of the subclass.

        Args:
            response: The response object to be processed, typically containing chunks of data.
        """
        ...

    @abstractmethod
    def extract_response_json(self, response: Response) -> dict:
        """
        Abstract method that must be implemented by the subclass.
        This method should provide logic to extract JSON data from the response, including headers and content.
        It is called after the response has been processed and is responsible for retrieving structured data
        that can be used by the application.

        Args:
            response: The response object from which to extract JSON data.
        """
        ...

    def create_streaming_response(
        self, token_streamer: Callable[[Send], Awaitable[None]]
    ) -> NBStreamingResponse:
        """
        Creates a streaming response using the provided token streamer.
        This method can be used by the subclass to create a response object that can be sent back to the client.
        The token streamer should be implemented to generate the content of the response according to the specific
        requirements of the subclass.

        Args:
            token_streamer: A callable that takes a send function and returns an awaitable. It's responsible for generating the content of the response.

        Returns:
            NBStreamingResponse: The streaming response object, ready to be sent to the client.
        """
        model_instance = NBStreamingResponseModel(token_streamer=token_streamer)

        return self.NBStreamingResponse(model_instance)

==> nimble-miner-api-0.0.16.tar.gz: nimble-miner-api-0.0.16/nimble/stream.py <==
# The MIT License (MIT)
# Copyright © 2024 Nimble Labs Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import nimble

from starlette.responses import StreamingResponse as _StreamingResponse
from starlette.responses import Response
from starlette.types import Send, Receive, Scope
from typing import Callable, Awaitable, List
from pydantic import BaseModel
from abc import ABC, abstractmethod


class NBStreamingResponseModel(BaseModel):
    """
    NBStreamingResponseModel is a Pydantic model that encapsulates the token streamer callable for Pydantic validation.
    It is used within the StreamingNucleon class to create a NBStreamingResponse object, which is responsible for handling
    the streaming of tokens.

    The token streamer is a callable that takes a send function and returns an awaitable. It is responsible for generating
    the content of the streaming response, typically by processing tokens and sending them to the client.

    This model ensures that the token streamer conforms to the expected signature and provides a clear interface for
    passing the token streamer to the NBStreamingResponse class.

    Attributes:
        token_streamer: Callable[[Send], Awaitable[None]]
            The token streamer callable, which takes a send function (provided by the ASGI server) and returns an awaitable.
            It is responsible for generating the content of the streaming response.
    """

    token_streamer: Callable[[Send], Awaitable[None]]


class StreamingNucleon(nimble.Nucleon, ABC):
    """
    The StreamingNucleon class is designed to be subclassed for handling streaming responses in the nimble network.
    It provides abstract methods that must be implemented by the subclass to deserialize, process streaming responses,
    and extract JSON data. It also includes a method to create a streaming response object.
    """

    class Config:
        validate_assignment = True

    class NBStreamingResponse(_StreamingResponse):
        """
        NBStreamingResponse is a specialized subclass of the Starlette StreamingResponse designed to handle the streaming
        of tokens within the nimble network. It is used internally by the StreamingNucleon class to manage the response
        streaming process, including sending headers and calling the token streamer provided by the subclass.

        This class is not intended to be directly instantiated or modified by developers subclassing StreamingNucleon.
        Instead, it is used by the create_streaming_response method to create a response object based on the token streamer
        provided by the subclass.

        Reference: https://www.starlette.io/responses/#streamingresponse
        """

        def __init__(self, model: NBStreamingResponseModel, **kwargs):
            """
            Initializes the NBStreamingResponse with the given token streamer model.

            Args:
                model: A NBStreamingResponseModel instance containing the token streamer callable, which is responsible
                    for generating the content of the response.
                **kwargs: Additional keyword arguments passed to the parent StreamingResponse class.
            """
            super().__init__(content=iter(()), **kwargs)
            self.token_streamer = model.token_streamer

        async def stream_response(self, send: Send):
            """
            Asynchronously streams the response by sending headers and calling the token streamer.

            This method is responsible for initiating the response by sending the appropriate headers, including the
            content type for event-streaming. It then calls the token streamer to generate the content and sends the
            response body to the client.

            Args:
                send: A callable to send the response, provided by the ASGI server.
            """
            headers = [
                (b"content-type", b"text/event-stream")
            ] + self.raw_headers

            await send(
                {
                    "type": "http.response.start",
                    "status": 200,
                    "headers": headers,
                }
            )

            await self.token_streamer(send)

            await send(
                {"type": "http.response.body", "body": b"", "more_body": False}
            )

        async def __call__(self, scope: Scope, receive: Receive, send: Send):
            """
            Asynchronously calls the stream_response method, allowing the NBStreamingResponse object to be used as an ASGI
            application.

            This method is part of the ASGI interface and is called by the ASGI server to handle the request and send the
            response. It delegates to the stream_response method to perform the actual streaming process.

            Args:
                scope: The scope of the request, containing information about the client, server, and request itself.
                receive: A callable to receive the request, provided by the ASGI server.
                send: A callable to send the response, provided by the ASGI server.
            """
            await self.stream_response(send)

    @abstractmethod
    async def process_streaming_response(self, response: Response):
        """
        Abstract method that must be implemented by the subclass.
        This method should provide logic to handle the streaming response, such as parsing and accumulating data.
        It is called as the response is being streamed from the network, and should be implemented to handle the specific
        streaming data format and requirements of the subclass.

        Args:
            response: The response object to be processed, typically containing chunks of data.
        """
        ...

    @abstractmethod
    def extract_response_json(self, response: Response) -> dict:
        """
        Abstract method that must be implemented by the subclass.
        This method should provide logic to extract JSON data from the response, including headers and content.
        It is called after the response has been processed and is responsible for retrieving structured data
        that can be used by the application.

        Args:
            response: The response object from which to extract JSON data.
        """
        ...

    def create_streaming_response(
        self, token_streamer: Callable[[Send], Awaitable[None]]
    ) -> NBStreamingResponse:
        """
        Creates a streaming response using the provided token streamer.
        This method can be used by the subclass to create a response object that can be sent back to the client.
        The token streamer should be implemented to generate the content of the response according to the specific
        requirements of the subclass.

        Args:
            token_streamer: A callable that takes a send function and returns an awaitable. It's responsible for generating the content of the response.

        Returns:
            NBStreamingResponse: The streaming response object, ready to be sent to the client.
        """
        model_instance = NBStreamingResponseModel(token_streamer=token_streamer)

        return self.NBStreamingResponse(model_instance)
